# Copyright © 2024 Gökdeniz Gülmez

import math

import mlx.core as mx
import mlx.nn as nn

from ..kan import KANLinear
from .convolution import kan_conv2d, multiple_convs_kan_conv2d

# Define a KAN convolution object
class KAN_Convolution(nn.Module):
    def __init__(
            self,
            kernel_size: tuple = (2,2),
            stride: tuple = (1,1),
            padding: tuple = (0,0),
            dilation: tuple = (1,1),
            grid_size: int = 5,
            spline_order: int = 3,
            scale_noise: float = 0.1,
            scale_base: float = 1.0,
            scale_spline: float = 1.0,
            hidden_act=nn.SiLU,
            grid_eps: float = 0.02,
            grid_range: tuple = [-1, 1]
        ):
        super(KAN_Convolution, self).__init__()
        self.grid_size = grid_size
        self.spline_order = spline_order
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation

        # Initialize KANLinear as convolution operation
        self.conv = KANLinear(
            in_features = math.prod(kernel_size),
            out_features = 1,
            grid_size=grid_size,
            spline_order=spline_order,
            scale_noise=scale_noise,
            scale_base=scale_base,
            scale_spline=scale_spline,
            hidden_act=hidden_act,
            grid_eps=grid_eps,
            grid_range=grid_range
        )

    def __call__(self, x: mx.array, update_grid=False):
        return kan_conv2d(x, self.conv, self.kernel_size[0], self.stride, self.dilation, self.padding)
    
    def regularization_loss(self, regularize_activation=1.0, regularize_entropy=1.0):
        return sum( layer.regularization_loss(regularize_activation, regularize_entropy) for layer in self.layers)

# Define a KAN convolutional layer with multiple convolutions
class KAN_Convolutional_Layer(nn.Module):
    def __init__(
            self,
            n_convs: int = 1,
            kernel_size: tuple = (2,2),
            stride: tuple = (1,1),
            padding: tuple = (0,0),
            dilation: tuple = (1,1),
            grid_size: int = 5,
            spline_order:int = 3,
            scale_noise:float = 0.1,
            scale_base: float = 1.0,
            scale_spline: float = 1.0,
            hidden_act=nn.SiLU,
            grid_eps: float = 0.02,
            grid_range: tuple = [-1, 1]
        ):
        """
        Kan Convolutional Layer with multiple convolutions
        
        Args:
            n_convs (int): Number of convolutions to apply
            kernel_size (tuple): Size of the kernel
            stride (tuple): Stride of the convolution
            padding (tuple): Padding of the convolution
            dilation (tuple): Dilation of the convolution
            grid_size (int): Size of the grid
            spline_order (int): Order of the spline
            scale_noise (float): Scale of the noise
            scale_base (float): Scale of the base
            scale_spline (float): Scale of the spline
            hidden_act (torch.nn.Module): Activation function
            grid_eps (float): Epsilon of the grid
            grid_range (tuple): Range of the grid
        """


        super(KAN_Convolutional_Layer, self).__init__()
        self.grid_size = grid_size
        self.spline_order = spline_order
        self.kernel_size = kernel_size
        self.dilation = dilation
        self.padding = padding
        self.convs = []
        self.n_convs = n_convs
        self.stride = stride


        # Create n_convs KAN_Convolution objects
        for _ in range(n_convs):
            self.convs.append(
                KAN_Convolution(
                    kernel_size= kernel_size,
                    stride = stride,
                    padding=padding,
                    dilation = dilation,
                    grid_size=grid_size,
                    spline_order=spline_order,
                    scale_noise=scale_noise,
                    scale_base=scale_base,
                    scale_spline=scale_spline,
                    hidden_act=hidden_act,
                    grid_eps=grid_eps,
                    grid_range=grid_range
                )
            )

    def __call__(self, x: mx.array, update_grid=False):
        # If there are multiple convolutions, apply them all
        if self.n_convs>1:
            return multiple_convs_kan_conv2d(x, self.convs, self.kernel_size[0], self.stride, self.dilation, self.padding)
        
        # If there is only one convolution, apply it
        return self.convs[0].forward(x)


# Copyright © 2024 Gökdeniz Gülmez

import os

from mlx.utils import tree_unflatten
import mlx.core as mx
import mlx.nn as nn

from kan.args import ModelArgs

from global_utils.utils import load_config

class KANLinear(nn.Module):
    def __init__(
        self,
        in_features,
        out_features,
        grid_size=5,
        spline_order=3,
        scale_noise=0.1,
        scale_base=1.0,
        scale_spline=1.0,
        enable_standalone_scale_spline=True,
        hidden_act=nn.SiLU,
        grid_eps=0.02,
        grid_range=[-1, 1],
    ):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.grid_size = grid_size
        self.spline_order = spline_order

        # Calculate the step size for the grid
        h = (grid_range[1] - grid_range[0]) / grid_size
        # Create a grid of points for the splines
        grid = (
            (mx.arange(-spline_order, grid_size + spline_order + 1) * h + grid_range[0]).reshape(-1, 1)  # Reshape to a column vector
        )
        # Tile the grid for each input feature
        self.grid = mx.tile(grid, (1, in_features))

        # Initialize weights for the base layer and the spline layer
        self.base_weight = mx.random.uniform(shape=(out_features, in_features))
        self.spline_weight = mx.random.uniform(shape=(out_features, in_features, grid_size + spline_order))

        # Initialize scaler for spline weights if standalone scaling is enabled
        if enable_standalone_scale_spline:
            self.spline_scaler = mx.random.uniform(shape=(out_features, in_features))

        # Save the scaling factors and activation function
        self.scale_noise = scale_noise
        self.scale_base = scale_base
        self.scale_spline = scale_spline
        self.enable_standalone_scale_spline = enable_standalone_scale_spline
        self.hidden_act = hidden_act()
        self.grid_eps = grid_eps

        self.reset_parameters()

    def reset_parameters(self):
        # Initialize base_weight with random values scaled by scale_base
        self.base_weight = mx.random.uniform(shape=self.base_weight.shape) * self.scale_base

        # Initialize spline_weight with random values scaled by scale_spline
        self.spline_weight = mx.random.uniform(shape=self.spline_weight.shape) * self.scale_spline

        if self.enable_standalone_scale_spline:
            # Initialize spline_scaler with random values scaled by scale_spline
            self.spline_scaler = mx.random.uniform(shape=self.spline_scaler.shape) * self.scale_spline


    def b_splines(self, x):
        # Ensure input x has the correct dimensions
        assert x.ndim == 2 and x.shape[1] == self.in_features

        x = x[:, :, None]  # Add a new dimension to x
        grid_reshaped = self.grid[:-1].T[None, :, :]  # Reshape to (1, in_features, grid_size)
        next_grid_reshaped = self.grid[1:].T[None, :, :]  # Reshape to (1, in_features, grid_size)

        # Compute the base splines
        bases = mx.logical_and(x >= grid_reshaped, x < next_grid_reshaped).astype(x.dtype)
        for k in range(1, self.spline_order + 1):
            # Update the splines for each order
            bases = (
                (x - self.grid[:-(k+1)].T[None, :, :]) /
                (self.grid[k:-1].T[None, :, :] - self.grid[:-(k+1)].T[None, :, :]) *
                bases[:, :, :-1]
            ) + (
                (self.grid[(k+1):].T[None, :, :] - x) /
                (self.grid[(k+1):].T[None, :, :] - self.grid[1:(-k)].T[None, :, :]) *
                bases[:, :, 1:]
            )
        # Ensure the bases have the correct shape
        assert bases.shape == (x.shape[0], self.in_features, self.grid_size + self.spline_order)
        return bases

        
    def curve2coeff(self, x, y):
        # Ensure input x and y have the correct dimensions
        assert x.ndim == 2 and x.shape[1] == self.in_features
        assert y.shape == (x.shape[0], self.in_features, self.out_features)
        
        # Transpose and reshape the splines and y for solving the least squares problem
        A = self.b_splines(x).transpose(1, 0, 2)
        B = y.transpose(1, 0, 2)  
        solution = mx.linalg.lstsq(A, B).solution
        result = solution.transpose(0, 2, 1)

        # Ensure the result has the correct shape
        assert result.shape == (
            self.out_features,
            self.in_features,  
            self.grid_size + self.spline_order,
        )
        return result


# loschen und commit
    @property
    def scaled_spline_weight(self):
        # Scale the spline weight if standalone scaling is enabled
        return self.spline_weight * (
            self.spline_scaler[:, :, None] 
            if self.enable_standalone_scale_spline
            else 1.0
        )

    def __call__(self, x):
        # Ensure input x has the correct dimensions
        assert x.ndim == 2 and x.shape[1] == self.in_features

        # Compute the base output
        base_output = mx.matmul(self.hidden_act(x), self.base_weight.T)

        # Compute the spline output
        spline_output = mx.matmul(
            self.b_splines(x.reshape(x.shape[0], -1)).reshape(x.shape[0], -1),
            self.scaled_spline_weight.reshape(self.out_features, -1).T,  
        )

        # Return the sum of the base output and the spline output
        return base_output + spline_output

    def update_grid(self, x, margin=0.01):
        # Ensure input x has the correct dimensions
        assert x.ndim == 2 and x.shape[1] == self.in_features
        batch = x.shape[0]
        
        # Compute splines and their outputs
        splines = self.b_splines(x)
        splines = splines.transpose(1, 0, 2)
        orig_coeff = self.scaled_spline_weight
        orig_coeff = orig_coeff.transpose(1, 2, 0) 
        unreduced_spline_output = mx.matmul(splines, orig_coeff)
        unreduced_spline_output = unreduced_spline_output.transpose(1, 0, 2)
        
        # Sort x and create adaptive and uniform grids
        x_sorted = mx.sort(x, axis=0)
        grid_adaptive = x_sorted[
            mx.linspace(
                0, batch - 1, self.grid_size + 1, dtype=mx.int32
            ).astype(mx.int64) 
        ]
        
        uniform_step = (x_sorted[-1] - x_sorted[0] + 2 * margin) / self.grid_size
        grid_uniform = (
            mx.arange(self.grid_size + 1).reshape(-1, 1).astype(mx.float32)
            * uniform_step
            + x_sorted[0] 
            - margin
        )
        
        # Blend the adaptive and uniform grids
        grid = self.grid_eps * grid_uniform + (1 - self.grid_eps) * grid_adaptive
        grid = mx.concatenate(
            [
                grid[:1] 
                - uniform_step
                * mx.arange(self.spline_order, 0, -1).reshape(-1, 1),
                grid,
                grid[-1:]
                + uniform_step 
                * mx.arange(1, self.spline_order + 1).reshape(-1, 1),
            ],
            axis=0,
        )
        
        # Update the grid and spline weights
        self.grid.set_value(grid.T)
        self.spline_weight.set_value(self.curve2coeff(x, unreduced_spline_output))
        
    def regularization_loss(self, regularize_activation=1.0, regularize_entropy=1.0):
        # Compute L1 norm of the spline weights
        l1_fake = mx.abs(self.spline_weight).mean(axis=-1)
        regularization_loss_activation = l1_fake.sum()

        # Compute entropy of the spline weights
        p = l1_fake / regularization_loss_activation
        regularization_loss_entropy = -mx.sum(p * mx.log(p))

        # Return the combined regularization loss
        return (
            regularize_activation * regularization_loss_activation
            + regularize_entropy * regularization_loss_entropy
        )

        
class KAN(nn.Module):
    def __init__(
        self,
        args: ModelArgs,
        layers_hidden=None,
    ):
        super().__init__()

        self.args = args

        if layers_hidden is None:
            if args.layers_hidden is None:
                layers_hidden = [args.in_features * args.out_features] + [args.hidden_dim] * (args.num_layers - 1) + [args.num_classes]
            else:
                layers_hidden = args.layers_hidden
        else:
            layers_hidden = layers_hidden
            self.args.layers_hidden = layers_hidden  # Update ModelArgs.layers_hidden

        # Save the grid and spline parameters
        self.grid_size = args.grid_size
        self.spline_order = args.spline_order
        
        self.layers = [] # Initialize the list of layers

        # Create KANLinear layers based on the hidden layers provided
        for in_features, out_features in zip(layers_hidden, layers_hidden[1:]):
            self.layers.append(
                KANLinear(
                    in_features,
                    out_features,
                    grid_size=args.grid_size,
                    spline_order=args.spline_order,
                    scale_noise=args.scale_noise,
                    scale_base=args.scale_base,
                    scale_spline=args.scale_spline,
                    hidden_act=args.hidden_act,
                    grid_eps=args.grid_eps,
                    grid_range=args.grid_range,
                )
            )
    
    def __call__(self, x, update_grid=False):
        # Pass the input x through each layer
        for layer in self.layers:
            if update_grid:
                layer.update_grid(x) # Update the grid if required
            x = layer(x) # Pass the input through the current layer
        return x
    
    def regularization_loss(self, regularize_activation=1.0, regularize_entropy=1.0):
        # Compute the regularization loss for all layers
        return mx.add(*(
            layer.regularization_loss(regularize_activation, regularize_entropy) 
            for layer in self.layers
        ))
    
    @staticmethod
    def load_model(folder_path: str):
        config_path = os.path.join(folder_path, "config.json")
        npz_model_path = os.path.join(folder_path, "model.safetensors")
        
        config = load_config(config_path)
        model = KAN(config)

        # Load weights and update model
        weights = tree_unflatten(list(mx.load(npz_model_path).items()))
        model.update(weights)
        
        return model


import argparse
import time
import logging
from tqdm import tqdm
import mlx.core as mx
import mlx.nn as nn
import mlx.optimizers as optim
import numpy as np
import quick_scripts.mnist as mnist

from kan import KAN
from kan.architectures.KANMLP import LlamaKANMLP, BasicKANMLP

from global_utils.utils import get_parameters, save_model, save_config, create_save_directory

def loss_fn(model, X, y):
    return mx.mean(nn.losses.cross_entropy(model(X), y))

def eval_fn(model, X, y):
    return mx.mean(mx.argmax(model(X), axis=1) == y)

def batch_iterate(batch_size, X, y):
    perm = mx.array(np.random.permutation(y.size))
    for s in range(0, y.size, batch_size):
        ids = perm[s : s + batch_size]
        yield X[ids], y[ids]

def batched_eval_fn(model: nn.Module, X: mx.array, y: mx.array, batch_size: int) -> float:
    total_correct = 0
    total_samples = 0
    for batch_X, batch_y in batch_iterate(batch_size, X, y):
        predictions = mx.argmax(model(batch_X), axis=1)
        total_correct += mx.sum(predictions == batch_y)
        total_samples += batch_y.size
    return total_correct / total_samples

def train_epoch_batched(model: nn.Module, optimizer: optim.Optimizer, train_images: mx.array, train_labels: mx.array, batch_size: int, loss_and_grad_fn, clip_grad_norm: bool = False) -> float:
    total_loss = 0
    num_batches = 0

    model.train()
    for X, y in tqdm(batch_iterate(batch_size, train_images, train_labels), total=len(train_images) // batch_size, desc='Training'):
        loss, grads = loss_and_grad_fn(model, X, y)
        if clip_grad_norm:
            optim.clip_grad_norm(grads, max_norm=1.0)
        optimizer.update(model, grads)
        mx.eval(model.parameters(), optimizer.state)
        total_loss += loss.item()
        num_batches += 1
    return total_loss / num_batches

def train_epoch(model: nn.Module, optimizer: optim.Optimizer, train_images: mx.array, train_labels: mx.array, batch_size: int, loss_and_grad_fn, clip_grad_norm: bool = False) -> float:
    model.train()
    for X, y in tqdm(batch_iterate(batch_size, train_images, train_labels), total=len(train_images) // batch_size, desc='Training'):
        loss, grads = loss_and_grad_fn(model, X, y)
        if mx.isnan(loss).any():
            raise ValueError("Encountered NaN in loss")
        if clip_grad_norm:
            optim.clip_grad_norm(grads, max_norm=1.0)
        optimizer.update(model, grads)
        mx.eval(model.parameters(), optimizer.state)
    return loss.item()

def main(args):
    seed = args.seed
    num_epochs = args.num_epochs
    learning_rate = args.learning_rate
    weight_decay = args.weight_decay

    np.random.seed(seed)

    print("\nLoading Dataset...")
    train_images, train_labels, test_images, test_labels = map(
        mx.array, getattr(mnist, args.dataset)()
    )
    print("...Dataset Loaded")

    print("\nInitializing and creating model...")
    # layers = [args.in_features * args.out_features] + [args.hidden_dim] * (args.num_layers - 1) + [args.num_classes]
    # model = KAN(layers)
    model = LlamaKANMLP(args.in_features, args.hidden_dim, args.out_features)
    print("Model initialized and created...")

    mx.eval(model.parameters())
    total_params = get_parameters(model)
    print(f"\nTraining a Kolmogorov–Arnold Network with {total_params} parameters")

    loss_and_grad_fn = nn.value_and_grad(model, loss_fn)
    optimizer = optim.AdamW(learning_rate=learning_rate, weight_decay=weight_decay)

    print(f"\nStarting to train model for {num_epochs} epochs")
    for e in range(args.num_epochs):
        current_epoch = e + 1
        tic = time.perf_counter()
        if args.train_batched:
            loss = train_epoch_batched(model, optimizer, train_images, train_labels, args.batch_size, loss_and_grad_fn=loss_and_grad_fn, clip_grad_norm=args.clip_grad_norm)
        else:
            loss = train_epoch(model, optimizer, train_images, train_labels, args.batch_size, loss_and_grad_fn=loss_and_grad_fn, clip_grad_norm=args.clip_grad_norm)
        toc = time.perf_counter()
        logging.info(f"Epoch {current_epoch}: Train Loss: {loss:.4f}, Time {toc - tic:.3f} (s)")
        print(f"Epoch {current_epoch}: Train Loss: {loss:.4f}, Time {toc - tic:.3f} (s)")

        if current_epoch == 1 or current_epoch % args.eval_report_count == 0 or e == args.num_epochs:
            tic = time.perf_counter()
            accuracy = batched_eval_fn(model, test_images, test_labels, args.batch_size)
            toc = time.perf_counter()
            logging.info(f"         Epoch {current_epoch}: Test accuracy {accuracy.item():.8f}, Time {toc - tic:.3f} (s)")
            print(f"         Epoch {current_epoch}: Test accuracy {accuracy.item():.8f}, Time {toc - tic:.3f} (s)")

    mx.eval(model.parameters(), optimizer.state)

    save_dir = create_save_directory(args.save_path)

    save_model(model=model, save_path=save_dir)
    save_config(args, save_dir)

if __name__ == "__main__":
    parser = argparse.ArgumentParser("Train KAN on MNIST with MLX.")
    parser.add_argument("--cpu", action="store_true", help="Use the CPU instead og Metal GPU backend.")
    parser.add_argument("--dataset", type=str, default="mnist", choices=["mnist", "fashion_mnist"], help="The dataset to use.")
    parser.add_argument("--num-layers", type=int, default=2, help="Number of layers in the model.")
    parser.add_argument("--hidden-dim", type=int, default=64, help="Number of hidden units in each layer.")
    parser.add_argument("--num-classes", type=int, default=10, help="Number of output classes.")
    parser.add_argument("--in-features", type=int, default=28, help="Number input features.")
    parser.add_argument("--out-features", type=int, default=28, help="Number output features.")
    parser.add_argument("--num-epochs", type=int, default=1, help="Number of epochs to train.")
    parser.add_argument("--batch-size", type=int, default=64, help="Batch size for training.")
    parser.add_argument("--learning-rate", type=float, default=1e-4, help="Learning rate for the optimizer.")
    parser.add_argument("--weight-decay", type=float, default=1e-4, help="Weight decay for the optimizer.")
    parser.add_argument("--eval-report-count", type=int, default=10, help="Number of epochs to report validations / test accuracy.")
    parser.add_argument("--seed", type=int, default=0, help="Random seed for reproducibility.")
    parser.add_argument("--train-batched", action="store_true", help="Train the model with batching.")
    parser.add_argument("--clip-grad-norm", action="store_true", help="Use gradient clipping.")
    parser.add_argument("--save-path", type=str, default="trained_kan_model", help="Directory to save the model and config.")
    args = parser.parse_args()

    if args.cpu:
        mx.set_default_device(mx.cpu)

    main(args)


from typing import Optional
from tqdm import tqdm
import time

import numpy as np

import mlx.optimizers as optimizer
import mlx.core as mx
import mlx.nn as nn

from trainer.trainer_args import TrainArgs
from global_utils.utils import get_parameters, save_model, save_config, create_save_directory


def loss_fn(model, X, y):
    return mx.mean(nn.losses.cross_entropy(model(X), y))

def eval_fn(model, X, y):
    return mx.mean(mx.argmax(model(X), axis=1) == y)


def batch_iterate(batch_size, X, y):
    perm = mx.array(np.random.permutation(y.size))
    for s in range(0, y.size, batch_size):
        ids = perm[s : s + batch_size]
        yield X[ids], y[ids]

def batched_eval_fn(model: nn.Module, X: mx.array, y: mx.array, batch_size: int) -> float:
    total_correct = 0
    total_samples = 0
    for batch_X, batch_y in batch_iterate(batch_size, X, y):
        predictions = mx.argmax(model(batch_X), axis=1)
        total_correct += mx.sum(predictions == batch_y)
        total_samples += batch_y.size
    return total_correct / total_samples


def train_epoch_batched(
        model: nn.Module,
        optimizer: optimizer.Optimizer,
        train_inputs: mx.array,
        train_labels: mx.array,
        batch_size: int,
        loss_and_grad_fn,
        clip_grad_norm: bool = False
    ) -> float:
    total_loss = 0
    num_batches = 0

    model.train()
    for X, y in tqdm(batch_iterate(batch_size, train_inputs, train_labels), total=len(train_inputs) // batch_size, desc='Training'):
        loss, grads = loss_and_grad_fn(model, X, y)
        if mx.isnan(loss).any():
            raise ValueError("Encountered NaN in loss")
        if clip_grad_norm:
            optimizer.clip_grad_norm(grads, max_norm=1.0)
        optimizer.update(model, grads)
        mx.eval(model.parameters(), optimizer.state)
        total_loss += loss.item()
        num_batches += 1
    return total_loss / num_batches


def train_step_batched(
        model: nn.Module,
        optimizer: optimizer.Optimizer,
        train_inputs: mx.array,
        train_labels: mx.array,
        batch_size: int,
        loss_and_grad_fn,
        clip_grad_norm: bool = False
    ) -> float:
    total_loss = 0
    num_batches = 0

    model.train()
    batch_iter = batch_iterate(batch_size, train_inputs, train_labels)
    try:
        X, y = next(batch_iter)
        loss, grads = loss_and_grad_fn(model, X, y)
        if mx.isnan(loss).any():
            raise ValueError("Encountered NaN in loss")
        if clip_grad_norm:
            optimizer.clip_grad_norm(grads, max_norm=1.0)
        optimizer.update(model, grads)
        mx.eval(model.parameters(), optimizer.state)
        total_loss += loss.item()
        num_batches += 1
    except StopIteration:
        pass

    return total_loss / num_batches


def SimpleTrainer(
        model: nn.Module,
        args: TrainArgs,
        train_set: Optional[tuple] = None,
        validation_set: Optional[tuple] = None,
        test_set: Optional[tuple] = None,
        validation_interval: Optional[int] = None,
        logging_interval: int = 10
    ):
    max_steps = args.max_steps
    epochs = args.epochs
    max_train_batch_size = args.max_train_batch_size
    max_val_batch_size = args.max_val_batch_size
    max_test_batch_size = args.max_test_batch_size
    learning_rate = args.learning_rate
    weight_decay = args.weight_decay
    clip_grad_norm = args.clip_grad_norm
    save_path = args.save_path

    train_images, train_labels = map(mx.array, train_set)
    validation_images, validation_labels = map(mx.array, validation_set)
    test_images, test_labels = map(mx.array, test_set)

    loss_and_grad_fn = nn.value_and_grad(model, loss_fn)
    opt = optimizer.AdamW(learning_rate=learning_rate, weight_decay=weight_decay)

    total_prams = get_parameters(model=model)
    print(f"\nStarting to train model with {total_prams} Params, for {epochs if max_steps == 0 else max_steps} {"epochs" if max_steps == 0 else "steps"}")
    
    if max_steps > 0:
        # Training using steps
        current_step = 0
        with tqdm(total=max_steps, desc="Training Steps") as pbar:
            while current_step < max_steps:
                tic = time.perf_counter()
                loss = train_step_batched(model, opt, train_images, train_labels, max_train_batch_size, loss_and_grad_fn=loss_and_grad_fn, clip_grad_norm=clip_grad_norm)
                toc = time.perf_counter()
                current_step += 1

                if current_step % logging_interval == 0:
                    pbar.set_postfix({"Train Loss": f"{loss:.4f}", "Time (s)": f"{toc - tic:.3f}"})
                pbar.update(1)

                if validation_interval and (current_step % validation_interval == 0 or current_step == max_steps):
                    tic = time.perf_counter()
                    val_accuracy = batched_eval_fn(model, validation_images, validation_labels, max_val_batch_size)
                    toc = time.perf_counter()
                    tqdm.write(f"-----Step {current_step}: Validation accuracy {val_accuracy.item():.8f}, Time {toc - tic:.3f} (s)")
    else:
        # Training using epochs
        for e in range(epochs):
            current_epoch = e + 1
            tic = time.perf_counter()
            loss = train_epoch_batched(model, opt, train_images, train_labels, max_train_batch_size, loss_and_grad_fn=loss_and_grad_fn, clip_grad_norm=clip_grad_norm)
            toc = time.perf_counter()


            if current_epoch % logging_interval == 0:
                pbar.set_postfix(f"Epoch {current_epoch}: Train Loss: {loss:.4f}, Time {toc - tic:.3f} (s)")
            pbar.update(1)

            if validation_interval and (current_epoch % validation_interval == 0 or current_epoch == epochs):
                tic = time.perf_counter()
                val_accuracy = batched_eval_fn(model, validation_images, validation_labels, max_val_batch_size)
                toc = time.perf_counter()
                tqdm.write(f"-----Epoch {current_epoch}: Validation accuracy {val_accuracy.item():.8f}, Time {toc - tic:.3f} (s)")

    # Evaluate on the test set
    print("Testing the trained model.")
    tic = time.perf_counter()
    test_accuracy = batched_eval_fn(model, test_images, test_labels, max_test_batch_size)
    toc = time.perf_counter()
    print(f"Test accuracy {test_accuracy.item():.8f}, Time {toc - tic:.3f} (s)")

    mx.eval(model.parameters(), opt.state)
    save_dir = create_save_directory(save_path)
    save_model(model=model, save_path=save_dir)
    save_config(model_args=model.args, save_path=save_dir, train_args=args)


# Copyright © 2024 Gökdeniz Gülmez

from typing import Tuple

import mlx.core as mx
import mlx.nn as nn

from kan import KANLinear
from kan.kan_conv.kan_convolution import KAN_Convolutional_Layer

class LlamaKANMLP(nn.Module):
    def __init__(
            self,
            in_features: int,
            hidden_dim: int,
            out_features: int,
            grid_size=5,
            spline_order=3,
            scale_noise=0.1,
            scale_base=1.0,
            scale_spline=1.0,
            enable_standalone_scale_spline=True,
            hidden_act=nn.SiLU,
            grid_eps=0.02,
            grid_range=[-1, 1],
            ):
        super().__init__()
        self.w1 = KANLinear(
            in_features=in_features,
            out_features=hidden_dim,
            grid_size=grid_size,
            spline_order=spline_order,
            scale_noise=scale_noise,
            scale_base=scale_base,
            scale_spline=scale_spline,
            enable_standalone_scale_spline=enable_standalone_scale_spline,
            hidden_act=hidden_act,
            grid_eps=grid_eps,
            grid_range=grid_range
        )
        self.w3 = KANLinear(
            in_features=in_features,
            out_features=hidden_dim,
            grid_size=grid_size,
            spline_order=spline_order,
            scale_noise=scale_noise,
            scale_base=scale_base,
            scale_spline=scale_spline,
            enable_standalone_scale_spline=enable_standalone_scale_spline,
            hidden_act=hidden_act,
            grid_eps=grid_eps,
            grid_range=grid_range
        )
        self.w2 = KANLinear(
            in_features=hidden_dim,
            out_features=out_features,
            grid_size=grid_size,
            spline_order=spline_order,
            scale_noise=scale_noise,
            scale_base=scale_base,
            scale_spline=scale_spline,
            enable_standalone_scale_spline=True,
            hidden_act=nn.Identity,
            grid_eps=grid_eps,
            grid_range=grid_range
        )
    def __call__(self, x):
        return self.w2(self.w1(x) * self.w3(x))

class SmallKANMLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_dim: int,
        out_features: int,
        grid_size=5,
        spline_order=3,
        scale_noise=0.1,
        scale_base=1.0,
        scale_spline=1.0,
        enable_standalone_scale_spline=True,
        hidden_act=nn.SiLU,
        grid_eps=0.02,
        grid_range=[-1, 1],
    ):
        super().__init__()
        self.layer1 = KANLinear(
            in_features=in_features,
            out_features=hidden_dim,
            grid_size=grid_size,
            spline_order=spline_order,
            scale_noise=scale_noise,
            scale_base=scale_base,
            scale_spline=scale_spline,
            enable_standalone_scale_spline=enable_standalone_scale_spline,
            hidden_act=hidden_act,
            grid_eps=grid_eps,
            grid_range=grid_range
        )
        self.layer2 = KANLinear(
            in_features=hidden_dim,
            out_features=out_features,
            grid_size=grid_size,
            spline_order=spline_order,
            scale_noise=scale_noise,
            scale_base=scale_base,
            scale_spline=scale_spline,
            enable_standalone_scale_spline=enable_standalone_scale_spline,
            hidden_act=hidden_act,
            grid_eps=grid_eps,
            grid_range=grid_range
        )

    def __call__(self, x):
        return self.layer2(self.layer1(x))

class MiddleKANMLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_dim: int,
        out_features: int,
        grid_size=5,
        spline_order=3,
        scale_noise=0.1,
        scale_base=1.0,
        scale_spline=1.0,
        enable_standalone_scale_spline=True,
        hidden_act=nn.SiLU,
        grid_eps=0.02,
        grid_range=[-1, 1],
    ):
        super().__init__()
        self.layer1 = KANLinear(
            in_features=in_features,
            out_features=hidden_dim,
            grid_size=grid_size,
            spline_order=spline_order,
            scale_noise=scale_noise,
            scale_base=scale_base,
            scale_spline=scale_spline,
            enable_standalone_scale_spline=enable_standalone_scale_spline,
            hidden_act=hidden_act,
            grid_eps=grid_eps,
            grid_range=grid_range
        )
        self.layer2 = KANLinear(
            in_features=hidden_dim,
            out_features=hidden_dim,
            grid_size=grid_size,
            spline_order=spline_order,
            scale_noise=scale_noise,
            scale_base=scale_base,
            scale_spline=scale_spline,
            enable_standalone_scale_spline=enable_standalone_scale_spline,
            hidden_act=hidden_act,
            grid_eps=grid_eps,
            grid_range=grid_range
        )
        self.layer3 = KANLinear(
            in_features=hidden_dim,
            out_features=out_features,
            grid_size=grid_size,
            spline_order=spline_order,
            scale_noise=scale_noise,
            scale_base=scale_base,
            scale_spline=scale_spline,
            enable_standalone_scale_spline=enable_standalone_scale_spline,
            hidden_act=hidden_act,
            grid_eps=grid_eps,
            grid_range=grid_range
        )

    def __call__(self, x):
        x = self.layer1(x) # (in_features -> hidden_dim)
        x = self.layer2(x) # (hidden_dim -> hidden_dim)
        x = self.layer3(x) # (hidden_dim -> out_features)
        return x

class BigKANMLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_dim: int,
        out_features: int,
        grid_size=5,
        spline_order=3,
        scale_noise=0.1,
        scale_base=1.0,
        scale_spline=1.0,
        enable_standalone_scale_spline=True,
        hidden_act=nn.SiLU,
        grid_eps=0.02,
        grid_range=[-1, 1],
    ):
        super().__init__()
        self.layer1 = KANLinear(
            in_features=in_features,
            out_features=hidden_dim,
            grid_size=grid_size,
            spline_order=spline_order,
            scale_noise=scale_noise,
            scale_base=scale_base,
            scale_spline=scale_spline,
            enable_standalone_scale_spline=enable_standalone_scale_spline,
            hidden_act=hidden_act,
            grid_eps=grid_eps,
            grid_range=grid_range
        )
        self.layer2 = KANLinear(
            in_features=hidden_dim,
            out_features=hidden_dim,
            grid_size=grid_size,
            spline_order=spline_order,
            scale_noise=scale_noise,
            scale_base=scale_base,
            scale_spline=scale_spline,
            enable_standalone_scale_spline=enable_standalone_scale_spline,
            hidden_act=hidden_act,
            grid_eps=grid_eps,
            grid_range=grid_range
        )
        self.layer3 = KANLinear(
            in_features=hidden_dim,
            out_features=hidden_dim,
            grid_size=grid_size,
            spline_order=spline_order,
            scale_noise=scale_noise,
            scale_base=scale_base,
            scale_spline=scale_spline,
            enable_standalone_scale_spline=enable_standalone_scale_spline,
            hidden_act=hidden_act,
            grid_eps=grid_eps,
            grid_range=grid_range
        )
        self.layer4 = KANLinear(
            in_features=hidden_dim,
            out_features=out_features,
            grid_size=grid_size,
            spline_order=spline_order,
            scale_noise=scale_noise,
            scale_base=scale_base,
            scale_spline=scale_spline,
            enable_standalone_scale_spline=enable_standalone_scale_spline,
            hidden_act=hidden_act,
            grid_eps=grid_eps,
            grid_range=grid_range
        )

    def __call__(self, x):
        return self.layer4(self.layer3(self.layer2(self.layer1(x))))



class KANC_MLP(nn.Module):
    def __init__(
            self,
            in_features: int = 625,
            hidden_dim: int = 256,
            out_features: int = 10,
            n_convs: int = 5,
            kernel_size: Tuple[int, int] = (3, 3)
        ):
        super().__init__()
        self.conv1 = KAN_Convolutional_Layer(n_convs=n_convs, kernel_size=kernel_size)
        self.conv2 = KAN_Convolutional_Layer(n_convs=5, kernel_size=kernel_size)
        self.pool1 = nn.MaxPool2d(kernel_size=(2, 2))
        self.fc1 = nn.Linear(in_features, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, out_features)

    def __call__(self, x):
        x = self.pool1(self.conv1(x))
        x = self.pool1(self.conv2(x))
        x = mx.flatten(x)
        x = nn.ReLU(self.fc1(x))
        x = self.fc2(x)
        return nn.log_softmax(x, axis=1)


from typing import Any
import os
import json
import mlx.core as mx
import mlx.nn as nn

from mlx.utils import tree_flatten

from kan.args import ModelArgs

def create_save_directory(base_path):
    os.makedirs(base_path, exist_ok=True)
    print(f"Using directory: {base_path}")
    return base_path

def get_parameters(model):
    def nparams(m):
        if isinstance(m, (nn.QuantizedLinear, nn.QuantizedEmbedding)):
            return m.weight.size * (32 // m.bits)
        return sum(v.size for _, v in tree_flatten(m.parameters()))

    leaf_modules = tree_flatten(
        model.leaf_modules(), is_leaf=lambda m: isinstance(m, nn.Module)
    )
    total_p = sum(nparams(m) for _, m in leaf_modules) / 10**6
    return total_p

def save_model(model: nn.Module, save_path):
    model_path = os.path.join(save_path, "model.safetensors")
    flattened_tree = tree_flatten(model.trainable_parameters())
    mx.save_safetensors(model_path, dict(flattened_tree))
    print(f"Saved model to {str(save_path)}")

def save_config(model_args: Any, save_path: str, train_args: Any = None):
    def is_json_serializable(value):
        try:
            json.dumps(value)
            return True
        except (TypeError, OverflowError):
            return False

    def serialize_dataclass(dataclass_instance):
        if not hasattr(dataclass_instance, "__dict__"):
            return dataclass_instance
        return {k: v if is_json_serializable(v) else str(v) for k, v in dataclass_instance.__dict__.items() if not k.startswith("__") and not callable(v)}

    config = serialize_dataclass(model_args)
    if train_args is not None:
        config.update(serialize_dataclass(train_args))
    
    config_path = os.path.join(save_path, "config.json")
    with open(config_path, 'w') as f:
        json.dump(config, f, indent=4)
    print(f"Configuration file saved to {config_path}")


def load_config(file_path: str) -> ModelArgs:
    with open(file_path, 'r') as f:
        config_dict = json.load(f)
    
    # Remove any keys that are not valid for ModelArgs
    valid_keys = set(ModelArgs.__dataclass_fields__.keys())
    config_dict = {k: v for k, v in config_dict.items() if k in valid_keys}
    
    return ModelArgs(**config_dict)


# Copyright © 2024 Gökdeniz Gülmez

import mlx.core as mx
import mlx.nn as nn

from kan import KANLinear
from kan.kan_convolution.kanConv import KAN_Convolutional_Layer

class KKAN_Convolutional_Network(nn.Module):
    def __init__(
        self,
        in_features=625,
        out_features=10,
        grid_size=10,
        spline_order=3,
        scale_noise=0.01,
        scale_base=1.0,
        scale_spline=1.0,
        enable_standalone_scale_spline=True,
        hidden_act=nn.SiLU,
        grid_eps=0.02,
        grid_range=[0,1],
    ):
        super().__init__()
        self.conv1 = KAN_Convolutional_Layer(
            n_convs = 5,
            kernel_size= (3,3)
        )

        self.conv2 = KAN_Convolutional_Layer(
            n_convs = 5,
            kernel_size = (3,3)
        )

        self.pool1 = nn.MaxPool2d(
            kernel_size=(2, 2)
        )

        self.kan1 = KANLinear(
            in_features=in_features,
            out_features=out_features,
            grid_size=grid_size,
            spline_order=spline_order,
            scale_noise=scale_noise,
            scale_base=scale_base,
            scale_spline=scale_spline,
            enable_standalone_scale_spline=enable_standalone_scale_spline,
            hidden_act=hidden_act,
            grid_eps=grid_eps,
            grid_range=grid_range
        )

    def __call__(self, x):
        x = self.pool1(self.conv1(x))
        x = mx.flatten(self.pool1(self.conv2(x)))
        x = self.kan1(x)
        return nn.log_softmax(x, axis=1)


from typing import Optional, List
from dataclasses import dataclass

import mlx.nn as nn


@dataclass
class ModelArgs:
    layers_hidden: Optional[List[int]] = None
    model_type: str = "KAN"
    num_layers: int = 2
    in_features: int = 28
    out_features: int = 28
    hidden_dim: int = 64
    num_classes: int = 10
    grid_size: int = 5
    spline_order: float = 3
    scale_noise: float = 0.1
    scale_base: float = 1.0
    scale_spline: float = 1.0
    hidden_act = nn.SiLU
    grid_eps: float = 0.02
    grid_range = [-1, 1]


# Copyright © 2024 Gökdeniz Gülmez

import numpy as np
from typing import Union, List, Tuple

import mlx.core as mx

def extract_patches(arr, patch_size):
    """ Extract patches of size patch_size from arr. """
    arr_shape = arr.shape
    patch_shape = (arr_shape[0] - patch_size[0] + 1, arr_shape[1] - patch_size[1] + 1, patch_size[0], patch_size[1])
    
    strides = arr.strides + (arr.strides[0], arr.strides[1])
    
    return np.lib.stride_tricks.as_strided(arr, shape=patch_shape, strides=strides)

# Calculate output dimensions for convolution
def calc_out_dims(matrix, kernel_side, stride, dilation, padding):
    # Ensure the matrix has 4 dimensions
    if len(matrix.shape) != 4:
        raise ValueError("Input matrix must have 4 dimensions (batch_size, n_channels, height, width)")

    batch_size, n_channels, n, m = matrix.shape
    h_out = np.floor((n + 2 * padding[0] - kernel_side - (kernel_side - 1) * (dilation[0] - 1)) / stride[0]).astype(int) + 1
    w_out = np.floor((m + 2 * padding[1] - kernel_side - (kernel_side - 1) * (dilation[1] - 1)) / stride[1]).astype(int) + 1
    b = [kernel_side // 2, kernel_side // 2]
    return h_out, w_out, batch_size, n_channels

# Perform multiple convolutions using KAN
def multiple_convs_kan_conv2d(
    matrix,
    kernels, 
    kernel_side,
    stride= (1, 1), 
    dilation= (1, 1), 
    padding= (0, 0)
) -> mx.array:
    """Makes a 2D convolution with the kernel over matrix using defined stride, dilation and padding along axes.

    Args:
        matrix (batch_size, colors, n, m]): 2D matrix to be convolved.
        kernels (List[function]): List of 2D kernels to apply.
        stride (Tuple[int, int], optional): Tuple of the stride along axes. With the `(r, c)` stride we move on `r` pixels along rows and on `c` pixels along columns on each iteration. Defaults to (1, 1).
        dilation (Tuple[int, int], optional): Tuple of the dilation along axes. With the `(r, c)` dilation we distancing adjacent pixels in kernel by `r` along rows and `c` along columns. Defaults to (1, 1).
        padding (Tuple[int, int], optional): Tuple with number of rows and columns to be padded. Defaults to (0, 0).

    Returns:
        np.ndarray: 2D Feature map, i.e. matrix after convolution.
    """

    def calc_out_dims(matrix, kernel_side, stride, dilation, padding):
        batch_size, n_channels, height, width = matrix.shape
        padded_height = height + 2 * padding[0]
        padded_width = width + 2 * padding[1]
        kernel_height = kernel_side * dilation[0] - (dilation[0] - 1)
        kernel_width = kernel_side * dilation[1] - (dilation[1] - 1)
        h_out = (padded_height - kernel_height) // stride[0] + 1
        w_out = (padded_width - kernel_width) // stride[1] + 1
        return h_out, w_out, batch_size, n_channels

    # Compute output dimensions
    h_out, w_out, batch_size, n_channels = calc_out_dims(matrix, kernel_side, stride, dilation, padding)
    n_convs = len(kernels)

    # Initialize output tensor
    matrix_out = np.zeros((batch_size, n_channels * n_convs, h_out, w_out))

    # Pad the input matrix
    matrix_padded = np.pad(matrix, ((0, 0), (0, 0), (padding[0], padding[0]), (padding[1], padding[1])), mode='constant')

    # Apply convolutional kernels
    for batch in range(batch_size):
        for channel in range(n_channels):
            for kern_idx, kernel in enumerate(kernels):
                for i in range(0, h_out):
                    for j in range(0, w_out):
                        # Calculate window start and end indices
                        row_start = i * stride[0]
                        row_end = row_start + kernel_side * dilation[0]
                        col_start = j * stride[1]
                        col_end = col_start + kernel_side * dilation[1]

                        # Extract the patch
                        patch = matrix_padded[batch, channel, row_start:row_end:dilation[0], col_start:col_end:dilation[1]]

                        # Apply kernel and store the result
                        conv_result = kernel(patch)
                        matrix_out[batch, channel * n_convs + kern_idx, i, j] = conv_result

    return matrix_out

# Perform a single KAN convolution
def kan_conv2d(
    matrix: Union[List[List[float]], np.ndarray],
    kernel, 
    kernel_side,
    stride= (1, 1), 
    dilation= (1, 1), 
    padding= (0, 0)
) -> np.ndarray:
    """Makes a 2D convolution with the kernel over matrix using defined stride, dilation and padding along axes.

    Args:
        matrix (batch_size, colors, n, m]): 2D matrix to be convolved.
        kernel (function): 2D odd-shaped matrix (e.g. 3x3, 5x5, 13x9, etc.).
        stride (Tuple[int, int], optional): Tuple of the stride along axes. With the `(r, c)` stride we move on `r` pixels along rows and on `c` pixels along columns on each iteration. Defaults to (1, 1).
        dilation (Tuple[int, int], optional): Tuple of the dilation along axes. With the `(r, c)` dilation we distancing adjacent pixels in kernel by `r` along rows and `c` along columns. Defaults to (1, 1).
        padding (Tuple[int, int], optional): Tuple with number of rows and columns to be padded. Defaults to (0, 0).

    Returns:
        np.ndarray: 2D Feature map, i.e. matrix after convolution.
    """

    def calc_out_dims(matrix, kernel_side, stride, dilation, padding):
        batch_size, n_channels, height, width = matrix.shape
        padded_height = height + 2 * padding[0]
        padded_width = width + 2 * padding[1]
        kernel_height = kernel_side * dilation[0] - (dilation[0] - 1)
        kernel_width = kernel_side * dilation[1] - (dilation[1] - 1)
        h_out = (padded_height - kernel_height) // stride[0] + 1
        w_out = (padded_width - kernel_width) // stride[1] + 1
        return h_out, w_out, batch_size, n_channels

    # Compute output dimensions
    h_out, w_out, batch_size, n_channels = calc_out_dims(matrix, kernel_side, stride, dilation, padding)

    # Initialize output tensor
    matrix_out = np.zeros((batch_size, n_channels, h_out, w_out))

    # Pad the input matrix
    matrix_padded = np.pad(matrix, ((0, 0), (0, 0), (padding[0], padding[0]), (padding[1], padding[1])), mode='constant')

    # Apply convolutional kernel
    for batch in range(batch_size):
        for channel in range(n_channels):
            for i in range(0, h_out):
                for j in range(0, w_out):
                    # Calculate window start and end indices
                    row_start = i * stride[0]
                    row_end = row_start + kernel_side * dilation[0]
                    col_start = j * stride[1]
                    col_end = col_start + kernel_side * dilation[1]

                    # Extract the patch
                    patch = matrix_padded[batch, channel, row_start:row_end:dilation[0], col_start:col_end:dilation[1]]

                    # Apply kernel and store the result
                    conv_result = kernel(patch)
                    matrix_out[batch, channel, i, j] = conv_result

    return matrix_out


from typing import Optional, Tuple
from dataclasses import dataclass

import mlx.nn as nn


@dataclass
class TrainArgs:
    train_algorithm: str = "simple"
    dataset: str = "custom"
    max_steps: int = 0
    epochs: int = 2
    max_train_batch_size: int = 32
    max_val_batch_size: int = 32
    max_test_batch_size: int = 32
    learning_rate: float = 1e-3
    weight_decay: float = 1e-5
    clip_grad_norm: bool = False
    save_path: str = "./models"


from .version import __version__

def __getattr__(name):
    if name in ("KANLinear", "KAN"):
        from .kan import KANLinear, KAN
        return locals()[name]
    if name in ("LlamaKANMLP", "SmallKANMLP", "MiddleKANMLP", "BigKANMLP"):
        from .architectures.KANMLP import LlamaKANMLP, SmallKANMLP, MiddleKANMLP, BigKANMLP
        return locals()[name]

__all__ = ["KANLinear", "KAN", "LlamaKANMLP", "SmallKANMLP", "MiddleKANMLP", "BigKANMLP"]


import sys
from pathlib import Path
from setuptools import find_packages, setup

# Get the project root directory
root_dir = Path(__file__).parent

# Add the package directory to the Python path
package_dir = root_dir / "kan"

# Read the requirements from the requirements.txt file
# First try to read from the root directory, if not found, try the package directory
requirements_file = root_dir / "requirements.txt"
if not requirements_file.exists():
    requirements_file = package_dir / "requirements.txt"

if requirements_file.exists():
    with open(requirements_file) as fid:
        requirements = [l.strip() for l in fid.readlines()]
else:
    print("Warning: requirements.txt not found. Proceeding without dependencies.")


# Import the version from the package
version = {}
with open(str(package_dir / "version.py")) as f:
    exec(f.read(), version)

# Setup configuration
setup(
    name="mlx-kan",
    version=version['__version__'],
    description="KAN: Kolmogorov–Arnold Networks on Apple silicon with MLX.",
    long_description=open(root_dir / "README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author_email="goekdenizguelmez@gmail.com",
    author="Gökdeniz Gülmez",
    url="https://github.com/Goekdeniz-Guelmez/mlx-kan",
    license="Apache-2.0",
    install_requires=requirements,
    packages=find_packages(),
    python_requires=">=3.8",
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Operating System :: MacOS :: MacOS X"
    ],
)


# Copyright © 2023 Apple Inc.
# Copied from the mlx-examples repo

import gzip
import os
import pickle
from urllib import request

import numpy as np


def mnist(
    save_dir="/tmp", base_url="https://raw.githubusercontent.com/fgnt/mnist/master/", filename="mnist.pkl"
):
    """
    Load the MNIST dataset in 4 tensors: train images, train labels,
    test images, and test labels.

    Checks `save_dir` for already downloaded data otherwise downloads.

    Download code modified from:
      https://github.com/hsjeong5/MNIST-for-Numpy
    """

    def download_and_save(save_file):
        filename = [
            ["training_images", "train-images-idx3-ubyte.gz"],
            ["test_images", "t10k-images-idx3-ubyte.gz"],
            ["training_labels", "train-labels-idx1-ubyte.gz"],
            ["test_labels", "t10k-labels-idx1-ubyte.gz"],
        ]

        mnist = {}
        for name in filename:
            out_file = os.path.join("/tmp", name[1])
            request.urlretrieve(base_url + name[1], out_file)
        for name in filename[:2]:
            out_file = os.path.join("/tmp", name[1])
            with gzip.open(out_file, "rb") as f:
                mnist[name[0]] = np.frombuffer(f.read(), np.uint8, offset=16).reshape(
                    -1, 28 * 28
                )
        for name in filename[-2:]:
            out_file = os.path.join("/tmp", name[1])
            with gzip.open(out_file, "rb") as f:
                mnist[name[0]] = np.frombuffer(f.read(), np.uint8, offset=8)
        with open(save_file, "wb") as f:
            pickle.dump(mnist, f)

    save_file = os.path.join(save_dir, filename)
    if not os.path.exists(save_file):
        download_and_save(save_file)
    with open(save_file, "rb") as f:
        mnist = pickle.load(f)

    def preproc(x):
        return x.astype(np.float32) / 255.0

    mnist["training_images"] = preproc(mnist["training_images"])
    mnist["test_images"] = preproc(mnist["test_images"])
    return (
        mnist["training_images"],
        mnist["training_labels"].astype(np.uint32),
        mnist["test_images"],
        mnist["test_labels"].astype(np.uint32),
    )


def fashion_mnist(save_dir="/tmp"):
    return mnist(
        save_dir,
        base_url="http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/",
        filename="fashion_mnist.pkl",
    )


if __name__ == "__main__":
    train_x, train_y, test_x, test_y = mnist()
    assert train_x.shape == (60000, 28 * 28), "Wrong training set size"
    assert train_y.shape == (60000,), "Wrong training set size"
    assert test_x.shape == (10000, 28 * 28), "Wrong test set size"
    assert test_y.shape == (10000,), "Wrong test set size"


# Copyright © 2024 Gökdeniz Gülmez

import mlx.core as mx
import mlx.nn as nn

from kan import KANLinear
from kan.kan_conv.KAN_Convolution import KAN_Convolutional_Layer

class KanGNN(nn.Module):
    def __init__(
            self,
            in_features: int,
            hidden_dim: int,
            out_features: int,
            grid_size: int = 5,
            layers_hidden: int = 6,
            use_bias: bool = False
        ):
        super().__init__()
        self.layers_hidden = layers_hidden
        
        self.ln1 = nn.Linear(in_features, hidden_dim, bias=use_bias)
        #self.ln1 = KANLayer(in_feat, hidden_dim, grid_size, addbias=use_bias)

        self.layers = []
        for i in range(layers_hidden):
            self.layers.append(KANLinear(hidden_dim, hidden_dim, grid_size, addbias=use_bias))

        self.layers.append(nn.Linear(hidden_dim, out_features, bias=False))
        #self.layers.append(KANLayer(hidden_dim, out_features, grid_size, addbias=False))

        # self.layers = []
        # self.layers.append(nn.Linear(in_features, hidden_dim, bias=use_bias))
        # for i in range(layers_hidden):
        #     self.layers.append(nn.Linear(hidden_dim, hidden_dim, bias=use_bias))
        # self.layers.append(nn.Linear(hidden_dim, out_features, bias=use_bias))

    
    def forward(self, x, adj):
        x = self.ln1(x)
        #x = self.ln1(spmm(adj, x))
        for layer in self.layers[:self.layers_hidden-1]:
            x = layer(spmm(adj, x))
            #x = layer(x)
        x = self.layers[-1](x)
            
        return x.log_softmax(dim=-1)